# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides access to SCADE installation information."""

import inspect
import os
from pathlib import Path
from typing import Dict

import scade_env


def get_scade_home() -> Path:
    """
    Get the SCADE installation directory.

    For example, ``C:/Program Files/ANSYS Inc/v232/SCADE``.
    """
    # when run from a SCADE environment, use the built-in environment variable SCADE
    # when the script is run through python.exe instead od scade.exe -script.
    # we derive the scade home directory from the location of the module scade_env
    # which is locate in <home>/SCADE/bin
    try:
        scade_home = Path(inspect.getfile(scade_env))
        return scade_home.parent.parent.parent
    except TypeError:
        # in a SCADE environment, scade_env is a built-in module
        # and SCADE is set to <home>/SCADE
        return Path(os.environ['SCADE']).parent


def get_scade_properties() -> Dict[str, str]:
    """Get the content of the properties in ``<home>/common/scade.properties`` as a dictionary."""
    scade_home = get_scade_home()
    # the file must exist for any release of SCADE using Python 3.7, at least until 2023 R2
    with (scade_home / 'common' / 'scade.properties').open() as f:
        tokens = [_.split('=', 1) for _ in f.read().split('\n') if _]
        d = {_[0]: _[1] for _ in tokens}
    return d


def get_scade_version() -> int:
    """Get the version of SCADE.

    For example, ``232`` for SCADE 2023 R2.
    """
    props = get_scade_properties()
    return int(props['SCADE_STUDIO_NUMBER'][:3])



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides accessors for expressions.

This module wraps the expressions derived from ``suite.Expression`` using the classes
corresponding to the XSCADE serialization, as exposed in the SCADE documentation:
``SCADE Suite Metamodels for Java API > 12. Scade Language Metamodels > Expressions``.
These classes allow accessing precisely each member of an expression.

Use the :py:func:`~ansys.scade.apitools.expr.access.accessor` function to create
an accessor from any SCADE Suite expression.
"""

from __future__ import annotations

from collections import namedtuple
from typing import Optional

import scade.model.suite as suite

from .predef import Eck


class Expression:
    """
    Provides the top-level abstract class for the expression accessors.

    Parameters
    ----------
    expression : suite.Expression
        Expression to wrap.
    """

    def __init__(self, expression: suite.Expression):
        """Store the wrapped expression."""
        self.expression = expression


class Present(Expression):
    """
    Provides the presence of a signal.

    The format is ``'<signal>``.

    See the :ref:`present <ex__present>` example.

    Parameters
    ----------
    expression :
        Signal expression to wrap.
    """

    def __init__(self, expression: suite.ExprId):
        """Initialize the instance from the Scade expression."""
        assert expression.reference and expression.reference.is_signal()
        super().__init__(expression)

    @property
    def signal(self) -> suite.LocalVariable:
        """Signal of the expression."""
        return self.expression.reference


class Last(Expression):
    """
    Provides the last of a local variable.

    The format is ``last '<variable>``.

    See the :ref:`last <ex__last>` example.

    Parameters
    ----------
    expression :
        Last variable expression to wrap.
    """

    def __init__(self, expression: suite.ExprId):
        """Initialize the instance from the Scade expression."""
        assert expression.last
        super().__init__(expression)

    @property
    def variable(self) -> suite.LocalVariable:
        """Local variable of the expression."""
        return self.expression.reference


class IdExpression(Expression):
    """
    Provides the constant, sensor, or local variable.

    The format is ``<path>``.

    See the :ref:`id_expression <ex__id_expression>` example.

    Parameters
    ----------
    expression :
        Reference expression to wrap.
    """

    def __init__(self, expression: suite.ExprId):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)

    @property
    def path(self) -> suite.ConstVar:
        """Local variable, sensor, or constant of the expression."""
        return self.expression.reference


class ConstValue(Expression):
    """
    Provides the literal value.

    The format is ``<value>``.

    See the :ref:`const_value <ex__const_value>` example.

    Parameters
    ----------
    expression :
        Literal to wrap.
    """

    def __init__(self, expression: suite.ConstValue):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)

    @property
    def value(self) -> str:
        """Literal value."""
        return self.expression.value


class TextExpression(Expression):
    """
    Provides an expression with a syntax error.

    The format is ``any text which can't compile``.

    Parameters
    ----------
    expression :
        Erroneous expression to wrap.
    """

    def __init__(self, expression: suite.ExprText):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)
        self._text = expression.text

    @property
    def text(self) -> str:
        """Erroneous text."""
        return self._text


class CallExpression(Expression):
    """
    Provides the abstract class for expression calls.

    There are three kinds of operator calls:

    * Predefined operators
    * User operators
    * Higher-order operators

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)
        self._code = Eck(expression.predef_opr)
        self._name = expression.inst_name

    @property
    def name(self) -> str:
        """Instance name of the expression."""
        return self._name

    @property
    def code(self) -> Eck:
        """Code of the predefined operator call."""
        return self._code


# Data operators
LabelledExpression = namedtuple('LabelledExpression', ['label', 'flow'])
"""
Element of a structure.

The format is ``<label>: <flow>``.
"""


class DataStructOp(CallExpression):
    """
    Provides for construction of a structure.

    The format is ``{ <label>: <expression>, ..., <label> : <expression> }``.

    See the :ref:`data_struct_op <ex__data_struct_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.BLD_STRUCT
        super().__init__(expression)
        self._data = [LabelledExpression(_.label.name, accessor(_)) for _ in expression.parameters]

    @property
    def data(self) -> list[LabelledExpression]:
        """Pairs (str, Expression) to build the structure."""
        return self._data


class DataArrayOp(CallExpression):
    """
    Provides for construction of an array.

    The format is ``[ <expression>, ..., <expression> ]``.

    See the :ref:`data_array_op <ex__data_array_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.BLD_VECTOR
        super().__init__(expression)
        self._data = [accessor(_) for _ in expression.parameters]

    @property
    def data(self) -> list[Expression]:
        """Values to build the array."""
        return self._data


class ArrayOp(CallExpression):
    """
    Provides the abstract class for array expressions.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) in {Eck.TRANSPOSE, Eck.SLICE, Eck.PRJ_DYN}
        super().__init__(expression)
        self._array = accessor(expression.parameters[0])

    @property
    def array(self) -> Expression:
        """Array operand of the expression."""
        return self._array


class TransposeOp(ArrayOp):
    """
    Provides for the transposition of an array.

    See the :ref:`transpose_op <ex__transpose_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.TRANSPOSE
        super().__init__(expression)
        self._dimensions = (accessor(expression.parameters[1]), expression.parameters[2])

    @property
    def dimensions(self) -> tuple[Expression, Expression]:
        """Dimensions to transpose."""
        return self._dimensions


class SliceOp(ArrayOp):
    """
    Provides the slice of an array.

    See the :ref:`slice_op <ex__slice_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.SLICE
        super().__init__(expression)
        self._from_index = accessor(expression.parameters[1])
        self._to_index = accessor(expression.parameters[2])

    @property
    def from_index(self) -> Expression:
        """Start index of the slice."""
        return self._from_index

    @property
    def to_index(self) -> Expression:
        """End index of the slice."""
        return self._to_index


class Label(Expression):
    """
    Provides the label of a projection.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ConstValue):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)
        self._name = expression.value

    @property
    def name(self) -> str:
        """Name of the label."""
        return self._name


class PrjDynOp(ArrayOp):
    """
    Provides the dynamic projection of an array.

    See the :ref:`prj_dyn_op <ex__prj_dyn_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.PRJ_DYN
        super().__init__(expression)
        # the indexes are either a label identifying a structure's
        # field or an index expression
        self._indexes = [
            Label(_)
            if isinstance(_, suite.ConstValue) and not _.value.isnumeric()
            else accessor(_)
            for _ in expression.parameters[1:-1]
        ]
        self._default = accessor(expression.parameters[-1])

    @property
    def indexes(self) -> list[Expression]:
        """Path of the projection."""
        return self._indexes

    @property
    def default(self) -> Expression:
        """Default value of the projection."""
        return self._default


class ListExpression(Expression):
    """
    Provides a group of expressions.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.Expression):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)
        self._items = [accessor(_) for _ in expression.parameters]

    @property
    def items(self) -> list[Expression]:
        """Expressions of the group."""
        return self._items


class FlowOp(CallExpression):
    """
    Provides an abstract class for flow expressions.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    flow :
        Expression operand.
    """

    def __init__(self, expression: suite.ExprCall, flow: suite.Expression):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)
        self._flow = accessor(flow)

    @property
    def flow(self) -> Expression:
        """Flow operand of the expression."""
        return self._flow


class ScalarToVectorOp(CallExpression):
    """
    Provides the vector from a flow and a size.

    The format is ``<flow> ^ <size>``.

    See the :ref:`scalar_to_vector_op <ex__scalar_to_vector_op>` example.

    Notes
    -----
    The design differs slightly from the meta-model. Because the input
    must be a group of flows, the :class:`~ScalarToVectorOp` class does not inherit
    from the :class:`~FlowOp` class. It exposes directly the list of input flows instead of
    having a flow that is a :class:`~ListExpression` class.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.SCALAR_TO_VECTOR
        super().__init__(expression)
        # the last parameter is the size
        self._flows = [accessor(_) for _ in expression.parameters[:-1]]
        self._size = accessor(expression.parameters[-1])

    @property
    def flows(self) -> list[Expression]:
        """Input flows."""
        return self._flows

    @property
    def size(self) -> Expression:
        """Size of the vector."""
        return self._size


class ProjectionOp(FlowOp):
    """
    Provides the abstract class for static projection/assignment of a flow.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    flow :
        Expression operand.
    with_expressions :
        Path of the expression.
    """

    def __init__(
        self,
        expression: suite.ExprCall,
        flow: suite.Expression,
        with_expressions: list[suite.Expression],
    ):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) in {Eck.PRJ, Eck.CHANGE_ITH}
        super().__init__(expression, flow)
        # the indexes are either a label identifying a structure's
        # field or an index expression
        self._with = [
            Label(_)
            if isinstance(_, suite.ConstValue) and not _.value.isnumeric()
            else accessor(_)
            for _ in with_expressions
        ]

    @property
    def with_(self) -> list[Expression]:
        """
        Path of the expression.

        The elements are either labels or indexes.
        """
        return self._with


class PrjOp(ProjectionOp):
    """
    Provides the static projection of a flow.

    See the example :ref:`prj_op <ex__prj_op>`.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.PRJ
        super().__init__(expression, expression.parameters[0], expression.parameters[1:])


class ChgIthOp(ProjectionOp):
    """
    Provides the static assignment of a flow.

    See the :ref:`chg_ith_op <ex__chg_ith_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.CHANGE_ITH
        super().__init__(expression, expression.parameters[0], expression.parameters[2:])
        self._value = accessor(expression.parameters[1])

    @property
    def value(self) -> Expression:
        """Value assigned."""
        return self._value


class MakeOp(CallExpression):
    """
    Provides the make of a structure.

    The format is ``(make <type>)(<flow>, ..., <flow>)``.

    See the :ref:`make_op <ex__make_op>` example.

    Notes
    -----
    The design differs slightly from the meta-model. This class derives from
    the :class:`~CallExpression` class.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.MAKE
        super().__init__(expression)
        self._flows = [accessor(_) for _ in expression.parameters[0].parameters]
        self._type = expression.parameters[1].reference

    @property
    def flows(self) -> list[Expression]:
        """Components of the structure."""
        return self._flows

    @property
    def type_(self) -> suite.NamedType:
        """Type to build."""
        return self._type


class FlattenOp(CallExpression):
    """
    Provides flattening of a structure.

    The format is ``(flatten <type>)(<flow>)``.

    See the :ref:`flatten_op <ex__flatten_op>` example.

    Notes
    -----
    The design differs slightly from the meta-model. This class derives from
    the :class:`~CallExpression` class.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.FLATTEN
        super().__init__(expression)
        self._flow = accessor(expression.parameters[0])
        self._type = expression.parameters[1].reference

    @property
    def flow(self) -> Expression:
        """Input flow."""
        return self._flow

    @property
    def type_(self) -> suite.NamedType:
        """Type to build."""
        return self._type


class AryOp(CallExpression):
    """
    Provides the abstract class for unary, binary, and nary operators.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)


class UnaryOp(AryOp):
    """
    Provides an expression with one operand.

    The format is ``<operator> <operand>``, where ``<operator>`` is one of
    these:

    * ``reverse``
    * ``-``, ``+``
    * ``not``
    * ``lnot``

    See the :ref:`unary_op <ex__unary_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert len(expression.parameters) == 1
        super().__init__(expression)
        self._operand = accessor(expression.parameters[0])

    @property
    def operand(self) -> Expression:
        """Operand of the operator."""
        return self._operand


class NAryOp(AryOp):
    """
    Provides an expression with two or more operands.

    The format is ``<operand> <operator> <operand> <operator> ... <operator> <operand>``,
    where ``<operator>`` is one of these:

    * ``@``
    * ``and``, ``or``, ``xor``
    * ``+``, ``*``
    * ``land``, ``lor``

    See the :ref:`n_ary_op <ex__n_ary_op>` example.

    Notes
    -----
    The ``xor`` operator is binary, but it is stored as a nary operator in the XSCADE files.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert len(expression.parameters) >= 2
        super().__init__(expression)
        self._operands = [accessor(_) for _ in expression.parameters]

    @property
    def operands(self) -> list[Expression]:
        """Operand of the operator."""
        return self._operands


class BinaryOp(NAryOp):
    """
    Provides an expression with two operands.

    The format is ``<operand> <operator> <operand>``, where ``<operator>`` is one of these:

    * ``-``, ``/``, ``mod``
    * ``<``, ``<=``, ``>``, ``>=``, ``=``, ``<>``
    * ``lxor``, ``lsl``, ``lsr``
    * ``times``

    See the :ref:`binary_op <ex__binary_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert len(expression.parameters) == 2
        super().__init__(expression)


class NumericCastOp(FlowOp):
    """
    Provides the numeric cast of a flow.

    See the :ref:`numeric_cast_op <ex__numeric_cast_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert len(expression.parameters) == 2 and isinstance(
            expression.parameters[1], suite.ExprType
        )
        super().__init__(expression, expression.parameters[0])
        self._type = expression.parameters[1].type

    @property
    def type_(self) -> suite.Type:
        """Conversion type."""
        return self._type


class SharpOp(CallExpression):
    """
    Provides a sharp expression with two or more flows.

    The format is ``#(<flow>, ..., <flow>)``.

    See the :ref:`sharp_op <ex__sharp_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert len(expression.parameters) >= 2
        super().__init__(expression)
        self._flows = [accessor(_) for _ in expression.parameters]

    @property
    def flows(self) -> list[Expression]:
        """Operands of the operator."""
        return self._flows


class IfThenElseOp(CallExpression):
    """
    Provides a vector from a flow and a size.

    The format is ``if <if> then <then>, ..., <then> else <else>, ..., <else>``.

    See the :ref:`if_then_else_op <ex__if_then_else_op>` example.

    Notes
    -----
    The design differs slightly from the meta-model. Because the then/else
    parts must be groups of flows, the :class:`~IfThenElseOp` class exposes directly
    the list of then/else flows instead of a flow that is an instance of the
    :class:`~ListExpression` class.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.IF
        super().__init__(expression)
        self._if = accessor(expression.parameters[0])
        self._then = [accessor(_) for _ in expression.parameters[1].parameters]
        self._else = [accessor(_) for _ in expression.parameters[2].parameters]

    @property
    def if_(self) -> Expression:
        """Condition of the expression."""
        return self._if

    @property
    def then(self) -> list[Expression]:
        """Flows when the condition is true."""
        return self._then

    @property
    def else_(self) -> list[Expression]:
        """Flows when the condition is false."""
        return self._else


class CaseOp(CallExpression):
    """
    Provides the case expression.

    Here is the format:

    .. code::

       ( case <switch> of ``
         | <pattern> :   <flow>
         | ...
         | <pattern> :   <flow>
         | _ :   <default>)

    See the :ref:`case_op <ex__case_op>` example.

    Notes
    -----
    The design differs slightly from the meta-model. The ``Case`` case used to
    implement the ``case`` collection is replaced by a tuple (``pattern``, ``flow``).

    A new property, ``default``, provides the optional default value.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.CASE
        super().__init__(expression)
        self._switch = accessor(expression.parameters[0])
        flows = [accessor(_) for _ in expression.parameters[1].parameters]
        patterns = [accessor(_) for _ in expression.parameters[2].parameters]
        self._default = flows.pop() if len(flows) > len(patterns) else None
        self._cases: list = [(pattern, flow) for pattern, flow in zip(patterns, flows)]

    @property
    def switch(self) -> Expression:
        """Selector of the case expression."""
        return self._switch

    @property
    def cases(self) -> list[tuple[Expression, Expression]]:
        """Pairs (:class:`~Expression`, :class:`~Expression`) to build the case."""
        return self._cases

    @property
    def default(self) -> Optional[Expression]:
        """Value to use as default when not ``None``."""
        return self._default


class InitOp(CallExpression):
    """
    Provides for the initialization of flows.

    The format is ``<init>, ..., <init> -> <flow>, ..., <flow>``.

    See the :ref:`init_op <ex__init_op>` example.

    Notes
    -----
    The design differs slightly from the meta-model. Because the inputs
    must be groups of flows, the :class:`~InitOp` class does not inherit
    from the :class:`~FlowOp` class. It exposes directly the lists of flows and initial
    values instead of having flows that are instances of the :class:`~ListExpression`
    class.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert len(expression.parameters) == 2 and Eck(expression.predef_opr) == Eck.FOLLOW
        super().__init__(expression)
        self._flows = [accessor(_) for _ in expression.parameters[0].parameters]
        self._inits = [accessor(_) for _ in expression.parameters[1].parameters]

    @property
    def flows(self) -> list[Expression]:
        """Flows after the first cycle."""
        return self._flows

    @property
    def inits(self) -> list[Expression]:
        """Initial values of the flows."""
        return self._inits


class PreOp(CallExpression):
    """
    Provides the previous value of flows.

    The format is ``pre <flow>, ..., <flow>``.

    See the :ref:`pre_op <ex__pre_op>` example.

    Notes
    -----
    The design differs slightly from the meta-model. Because the input
    must be a group of flows, the :class:`~PreOp` class does not inherit
    from the :class:`~FlowOp` class. It exposes directly the lists of flows instead of
    having a flow that is an instance of a :class:`~ListExpression` class.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert Eck(expression.predef_opr) == Eck.PRE
        super().__init__(expression)
        self._flows = [accessor(_) for _ in expression.parameters]

    @property
    def flows(self) -> list[Expression]:
        """Flows after the first cycle."""
        return self._flows


class FbyOp(CallExpression):
    """
    Provides the delay of flows.

    The format is ``fby(<flow>, ..., <flow>; <delay>; <init>, ..., <init>)``.

    See the :ref:`fby_op <ex__fby_op>` example.

    Parameters
    ----------
    expression :
        Call expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert len(expression.parameters) >= 3 and Eck(expression.predef_opr) == Eck.FBY
        super().__init__(expression)
        n = int((len(expression.parameters) - 1) / 2)
        self._flows = [accessor(_) for _ in expression.parameters[:n]]
        self._delay = accessor(expression.parameters[n])
        self._inits = [accessor(_) for _ in expression.parameters[n + 1 :]]

    @property
    def flows(self) -> list[Expression]:
        """Flows after the first cycle."""
        return self._flows

    @property
    def delay(self) -> Expression:
        """Delay of the expression."""
        return self._delay

    @property
    def inits(self) -> list[Expression]:
        """Initialization values of the flows."""
        return self._inits


class OpCall(CallExpression):
    """
    Calls a user-defined operator.

    The format is ``<operator><< <instance parameter>, ...>>(<call parameter>, ...)``.

    Notes
    -----
    The design differs slightly from the meta-model. The :class:`~OpCall` class
    is no longer aggregated by the :class:`~CallExpression` class but derives from it.
    This leads to a simpler design.

    Parameters
    ----------
    expression :
        Expression to wrap.
    """

    def __init__(self, expression: suite.ExprCall):
        """Initialize the instance from the Scade expression."""
        assert expression.operator
        super().__init__(expression)
        self._call_parameters = [accessor(_) for _ in expression.parameters]
        self._instance_parameters = [accessor(_) for _ in expression.inst_parameters]
        self._operator = expression.operator

    @property
    def operator(self) -> suite.Operator:
        """User operator."""
        return self._operator

    @property
    def call_parameters(self) -> list[Expression]:
        """Call parameters."""
        return self._call_parameters

    @property
    def instance_parameters(self) -> list[Expression]:
        """Instance call parameters."""
        return self._instance_parameters


class OpOp(CallExpression):
    """
    Provides the abstract class for higher-order operators.

    Notes
    -----
    The design differs slightly from the meta-model. The :class:`~OpOp` class is
    no longer aggregated by the :class:`~CallExpression` class but derives from it.
    This leads to a simpler design.

    Parameters
    ----------
    expression :
        Higher-order expression to wrap.
    operator :
        Operator call expression.
    """

    def __init__(self, expression: suite.ExprCall, operator: CallExpression):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression)
        self._operator = operator

    @property
    def operator(self) -> CallExpression:
        """Underlying expression."""
        return self._operator


class ConditionalOp(OpOp):
    """
    Provides the abstract class for restart and activate operators.

    Parameters
    ----------
    expression :
        Higher-order expression to wrap.
    operator :
        Operator call expression.
    """

    def __init__(self, expression: suite.ExprCall, operator: CallExpression):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression, operator)
        self._every = accessor(expression.parameters[0])

    @property
    def every(self) -> Expression:
        """Condition expression."""
        return self._every


class RestartOp(ConditionalOp):
    """
    Provides for restart of an operator.

    See the :ref:`restart_op <ex__restart_op>` example.

    Parameters
    ----------
    expression :
        Higher-order expression to wrap.
    operator :
        Operator call expression.
    """

    pass


class CondactOp(ConditionalOp):
    """
    Provides the abstract class for activate operators.

    Notes
    -----
    The design differs slightly from the meta-model. Because the flow
    ``default`` must be a group of flows, the :class:`~CondactOp` class exposes directly
    the list of default values instead of an instance of the :class:`~ListExpression` class.

    Parameters
    ----------
    expression :
        Higher-order expression to wrap.
    operator :
        Operator call expression.
    """

    def __init__(self, expression: suite.ExprCall, operator: CallExpression):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression, operator)
        self._defaults = [accessor(_) for _ in expression.parameters[1].parameters]

    @property
    def defaults(self) -> list[Expression]:
        """Initialization or default values."""
        return self._defaults


class ActivateOp(CondactOp):
    """
    Activation of an operator with initial values.

    See the :ref:`activate_op <ex__activate_op>` example.

    Parameters
    ----------
    expression :
        Higher-order expression to wrap.
    operator :
        Operator call expression.
    """

    pass


class ActivateNoInitOp(CondactOp):
    """
    Provides activation of an operator with default values.

    See the :ref:`activate_no_init_op <ex__activate_no_init_op>` example.

    Parameters
    ----------
    expression :
        Higher-order expression to wrap.
    operator :
        Operator call expression.
    """

    pass


class IteratorOp(OpOp):
    """
    Provides the base class for iteration operators.

    See the :ref:`iterator_op <ex__iterator_op>` example.

    Parameters
    ----------
    expression :
        Higher-order expression to wrap.
    operator :
        Operator call expression.
    """

    def __init__(self, expression: suite.ExprCall, operator: CallExpression):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression, operator)
        self._size = accessor(expression.parameters[0])
        if self.code in {Eck.MAPFOLD, Eck.MAPFOLDI, Eck.MAPFOLDW, Eck.MAPFOLDWI}:
            self._accumulator_count = accessor(expression.parameters[1])
        else:
            # n/a
            self._accumulator_count = None

    @property
    def size(self) -> Expression:
        """Size of the iterator."""
        return self._size

    # TODO CREATE: mapfold with N accumulators...
    @property
    def accumulator_count(self) -> Optional[Expression]:
        """Number of accumulators when suitable, otherwise ``None``."""
        return self._accumulator_count


class PartialIteratorOp(IteratorOp):
    """
    Provides partial iteration of an operator.

    See the :ref:`partial_iterator_op <ex__partial_iterator_op>` example.

    Parameters
    ----------
    expression :
        Higher-order expression to wrap.
    operator :
        Operator call expression.
    """

    def __init__(self, expression: suite.ExprCall, operator: CallExpression):
        """Initialize the instance from the Scade expression."""
        super().__init__(expression, operator)
        # mapfold iterators have a extra parameter, second position which
        # defines the number of accumulators
        offset = 1 if self.code in {Eck.MAPFOLDW, Eck.MAPFOLDWI} else 0
        self._if = accessor(expression.parameters[1 + offset])
        # following might be empty when not MAP*W*
        if len(expression.parameters) == 3 + offset:
            assert self.code in {Eck.MAPW, Eck.MAPWI, Eck.MAPFOLDW, Eck.MAPFOLDWI}
            self._defaults = [accessor(_) for _ in expression.parameters[2 + offset].parameters]
        else:
            assert self.code in {Eck.FOLDW, Eck.FOLDWI}
            self._defaults = None

    @property
    def if_(self) -> Expression:
        """Condition of the iterator."""
        return self._if

    @property
    def defaults(self) -> Optional[list[Expression]]:
        """Default values when suitable, otherwise ``None``."""
        return self._defaults


UNARY_OPS = {
    Eck.REVERSE: 'reverse',
    Eck.NEG: '-',
    Eck.POS: '+',
    Eck.NOT: 'not',
    Eck.LNOT: 'lnot',
}
BINARY_OPS = {
    Eck.SUB: '-',
    Eck.SLASH: '/',
    Eck.MOD: 'mod',
    Eck.LESS: '<',
    Eck.LEQUAL: '<=',
    Eck.GREAT: '>',
    Eck.GEQUAL: '>=',
    Eck.EQUAL: '=',
    Eck.NEQUAL: '<>',
    Eck.LXOR: 'lxor',
    Eck.LSL: 'lsl',
    Eck.LSR: 'lsr',
    Eck.TIMES: 'times',
    # TODO CREATE: bitwise...
}
NARY_OPS = {
    Eck.CONCAT: '@',
    Eck.AND: 'and',
    Eck.OR: 'or',
    # xor shall be binary but stored as nary in the xscade files
    Eck.XOR: 'xor',
    Eck.PLUS: '+',
    Eck.MUL: '*',
    Eck.LAND: 'land',
    Eck.LOR: 'lor',
}
MAP_OPERATORS = {
    Eck.BLD_STRUCT: DataStructOp,
    Eck.BLD_VECTOR: DataArrayOp,
    Eck.TRANSPOSE: TransposeOp,
    Eck.SLICE: SliceOp,
    Eck.PRJ_DYN: PrjDynOp,
    Eck.SEQ_EXPR: ListExpression,
    Eck.SCALAR_TO_VECTOR: ScalarToVectorOp,
    Eck.PRJ: PrjOp,
    Eck.CHANGE_ITH: ChgIthOp,
    Eck.MAKE: MakeOp,
    Eck.FLATTEN: FlattenOp,
    Eck.NUMERIC_CAST: NumericCastOp,
    Eck.SHARP: SharpOp,
    Eck.IF: IfThenElseOp,
    Eck.CASE: CaseOp,
    Eck.FOLLOW: InitOp,
    Eck.PRE: PreOp,
    Eck.FBY: FbyOp,
}
MAP_HIGHER_ORDER = {
    Eck.RESTART: RestartOp,
    Eck.ACTIVATE: ActivateOp,
    Eck.ACTIVATE_NOINIT: ActivateNoInitOp,
    Eck.MAP: IteratorOp,
    Eck.MAPI: IteratorOp,
    Eck.MAPW: PartialIteratorOp,
    Eck.MAPWI: PartialIteratorOp,
    Eck.FOLD: IteratorOp,
    Eck.FOLDI: IteratorOp,
    Eck.FOLDW: PartialIteratorOp,
    Eck.FOLDWI: PartialIteratorOp,
    Eck.MAPFOLD: IteratorOp,
    Eck.MAPFOLDI: IteratorOp,
    Eck.MAPFOLDW: PartialIteratorOp,
    Eck.MAPFOLDWI: PartialIteratorOp,
}

# overall map of operators
map_operators = {}
map_operators.update(MAP_OPERATORS)
map_operators.update({_: UnaryOp for _ in UNARY_OPS})
map_operators.update({_: BinaryOp for _ in BINARY_OPS})
map_operators.update({_: NAryOp for _ in NARY_OPS})


def _modifier_accessor(modifier: suite.Expression, call: CallExpression) -> CallExpression:
    if modifier:
        call = _modifier_accessor(modifier.modifier, call)
        class_ = MAP_HIGHER_ORDER.get(Eck(modifier.predef_opr))
        if not class_:
            raise ValueError('Higher order operator %d not supported' % modifier.predef_opr)
        else:
            call = class_(modifier, call)

    return call


def accessor(expression: suite.Expression) -> Expression:
    """
    Build the accessor for a SCADE Suite expression.

    Parameters
    ----------
    expression : suite.Expression
        Expression to wrap.

    Returns
    -------
    Expression
        Expression accessor.
    """
    # ExprId
    if isinstance(expression, suite.ExprId):
        if expression.reference and expression.reference.is_signal():
            return Present(expression)
        elif expression.last:
            return Last(expression)
        else:
            return IdExpression(expression)

    # ConstValue
    if isinstance(expression, suite.ConstValue):
        return ConstValue(expression)

    # TextExpression
    if isinstance(expression, suite.ExprText):
        return TextExpression(expression)

    # ...
    if isinstance(expression, suite.ExprType):
        raise ValueError('Type sub-expressions not supported')

    assert isinstance(expression, suite.ExprCall)
    if expression.operator:
        call = OpCall(expression)
    else:
        class_ = map_operators.get(Eck(expression.predef_opr))
        if class_:
            call = class_(expression)
        else:
            raise ValueError('Predefined operator %d not supported' % expression.predef_opr)

    return _modifier_accessor(expression.modifier, call)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides helpers for creating expression trees.

Expression trees are intermediate structures to declare any arbitrary complex
expressions. They create the corresponding SCADE Suite expressions in the
context of a model element, such as the right part of an equation or the
default value of an output.

This module provides functions to create an expression tree for any expression
of the Scade language, including higher-order constructs. Thus, the intermediate
structures or classes defining the expression trees can be opaque.

Notes: The typing is relaxed in this module to ease the constructs.

* ``ET`` is an alias for ``ExpressionTree`` to shorten the declarations.

* ``EX``, which stands for extended expression tree, is defined as follows::

     Union[bool, int, float, str, suite.ConstVar, suite.NamedType, ET]

  This enhances the usability of these functions by accepting some values,
  such as Python literals, string values, or SCADE Python objects, as valid
  expression trees.

* ``LX``, which stands for extended lists of expression trees, is defined as follows::

     Union[EX, List[EX]]

  When the expressions accept an arbitrary number of input flows, such as
  if-then-else or fby, you can provide either one expression tree or a list
  of expression trees.

"""

from typing import List, Tuple, Union

import scade.model.suite as suite

from ..expr import Eck
from .project import _check_object
from .scade import _add_pending_link


# expression trees
class ExpressionTree:
    """Provides the top-level abstract class for expression trees."""

    def __init__(self, label: str = ''):
        """Any expression can have a label."""
        self.label = label

    def _build_expression(self, context: suite.Object) -> suite.Expression:
        """Build a SCADE Suite expression from the expression tree."""
        # must be overridden
        assert False  # pragma no cover

    def _set_label(self, expr: suite.Expression, context: suite.Object):
        """Add the label to the expression, if any."""
        if self.label:
            label = suite.Label(context)
            label.name = self.label
            expr.label = label


ET = ExpressionTree
"""Short name for an ``ExpressionTree`` instance to simplify the declarations."""

EX = Union[bool, int, float, str, suite.ConstVar, suite.NamedType, ET]
"""Extended expression tree to simplify use of the create functions."""

LX = Union[EX, List[EX]]
"""Extended lists of expression trees to simply the use of create functions."""


class _Value(ET):
    """Provides a literal value."""

    def __init__(self, value: str, kind: str, **kwargs):
        """Literal value."""
        super().__init__(**kwargs)
        self.value = value
        self.kind = kind

    def _build_expression(self, context: suite.Object) -> suite.Expression:
        """Build a SCADE Suite expression from the expression tree."""
        value = suite.ConstValue(context)
        value.value = self.value
        value.kind = self.kind
        # no comment...
        if self.kind == 'String':
            value.role_kind = 'For Structure'
        self._set_label(value, context)
        return value


class _Reference(ET):
    """Provides a reference to a SCADE constant or variable."""

    def __init__(self, reference: suite.ConstVar, **kwargs):
        """Initialize a constant, sensor, or local variable."""
        super().__init__(**kwargs)
        self.reference = reference

    def _build_expression(self, context: suite.Object) -> suite.Expression:
        """Build a SCADE Suite expression from the expression tree."""
        expr = suite.ExprId(context)
        self._set_label(expr, context)
        _add_pending_link(expr, 'reference', self.reference)
        return expr


class _Type(ET):
    """Creates a reference to a type."""

    def __init__(self, type_: suite.NamedType, **kwargs):
        """Initialize a named type."""
        super().__init__(**kwargs)
        self.type = type_

    def _build_expression(self, context: suite.Object) -> suite.Expression:
        """Build a SCADE Suite expression from the expression tree."""
        expr = suite.ExprType(context)
        _add_pending_link(expr, 'type', self.type)
        return expr


class _Call(ET):
    """Provides the base class for expression calls."""

    def __init__(
        self, args: List[ET], inst_args: List[ET], modifiers: List[ET], name: str = '', **kwargs
    ):
        super().__init__(**kwargs)
        self.args = args
        self.inst_args = inst_args
        self.modifiers = modifiers
        self.name = name

    def _build_expression(self, context: suite.Object) -> suite.Expression:
        """Build a SCADE Suite expression from the expression tree."""
        expr = suite.ExprCall()
        expr.inst_name = self.name
        expr.parameters = [_._build_expression(context) for _ in self.args]
        expr.inst_parameters = [_._build_expression(context) for _ in self.inst_args]
        owner = expr
        for modifier in self.modifiers:
            subexpr = modifier._build_expression(context)
            owner.modifier = subexpr
            owner = subexpr
        # label, even if very unlinkely
        self._set_label(expr, context)
        return expr


class _Predefined(_Call):
    """Provides a calls to a predefined operator."""

    def __init__(self, eck: Eck, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eck = eck

    def _build_expression(self, context: suite.Object) -> suite.Expression:
        """Build a SCADE Suite expression from the expression tree."""
        expr = super()._build_expression(context)
        expr.predef_opr = self.eck.value
        return expr


class _Operator(_Call):
    """Provides a call to a user operator."""

    def __init__(self, operator: suite.Operator, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.operator = operator

    def _build_expression(self, context: suite.Object) -> suite.Expression:
        """Build a SCADE Suite expression from the expression tree."""
        expr = super()._build_expression(context)
        expr.operator = self.operator
        return expr


def _normalize_tree(any: EX) -> ET:
    """Create expression tree instances from literals or SCADE objects."""
    if isinstance(any, ET):
        return any
    # collections
    if isinstance(any, tuple) or isinstance(any, list):
        if len(any) == 0:
            raise EmptyTreeError()
        return [_normalize_tree(_) for _ in any]
    # SCADE objects
    if isinstance(any, suite.ConstVar):
        return _Reference(any)
    if isinstance(any, suite.NamedType):
        return _Type(any)
    # literals
    # true | false | <integer> | <real> | ' <char> ' | " <ident> "
    if isinstance(any, bool):
        return _Value(str(any).lower(), 'Bool')
    elif isinstance(any, int):
        return _Value(str(any), 'Int')
    elif isinstance(any, float):
        return _Value(str(any), 'Real')
    elif isinstance(any, str):
        if _is_bool(any):
            return _Value(str(any), 'Bool')
        elif _is_int(any):
            # number with an optional suffix _i8, ui32, etc.
            return _Value(any, 'Int')
        elif _is_real(any):
            # number with an optional suffix _f32 or _f64
            return _Value(any, 'Real')
        elif len(any) > 1 and any[0] == "'":
            # no additional verification on the syntax
            return _Value(any, 'Char')
        elif any.isidentifier():
            # used in projections only
            return _Value(any, 'String')

    # fall through
    raise ExprSyntaxError('_normalize_tree', any)


def _normalize_tree_ex(any: LX) -> List[ET]:
    """Normalize a collection of expression trees or a single one."""
    if isinstance(any, list):
        return _normalize_tree(any)
    else:
        return [_normalize_tree(any)]


def _build_expression(tree: EX, context: suite.Object) -> suite.Expression:
    """
    Build an expression from an extended expression tree.

    Parameters
    ----------
    tree : EX
        Operand: expression tree.
    context : suite.Object
        Context of the creation of the expression.

    Returns
    -------
    suite.Expression
    """
    if tree is None:
        return None
    tree = _normalize_tree(tree)
    return tree._build_expression(context)


# association tables
_unary_ops = {
    '-': Eck.NEG,
    '+': Eck.POS,
    '!': Eck.NOT,
    'int': Eck.REAL2INT,
    'real': Eck.INT2REAL,
    'lnot': Eck.LNOT,
}
_binary_ops = {
    '&': Eck.AND,
    '|': Eck.OR,
    '^': Eck.XOR,
    '#': Eck.SHARP,
    '+': Eck.PLUS,
    '-': Eck.SUB,
    '*': Eck.MUL,
    '/': Eck.SLASH,
    ':': Eck.DIV,
    '%': Eck.MOD,
    '<': Eck.LESS,
    '<=': Eck.LEQUAL,
    '>': Eck.GREAT,
    '>=': Eck.GEQUAL,
    '=': Eck.EQUAL,
    '<>': Eck.NEQUAL,
    'land': Eck.LAND,
    'lor': Eck.LOR,
    'lxor': Eck.LXOR,
    '<<': Eck.LSL,
    '>>': Eck.LSR,
}
_nary_ops = {'&': Eck.AND, '|': Eck.OR, '^': Eck.XOR, '#': Eck.SHARP, '+': Eck.PLUS, '*': Eck.MUL}


def create_call(operator: suite.Operator, args: LX, inst_args: LX = None) -> ET:
    """
    Return the expression tree for a call to an operator.

    Parameters
    ----------
    op : suite.Operator
        Operator to call.
    args : Union[EX, List[EX]]
        Parameters: expression trees.
    inst_args : Union[EX, List[EX]]
        Instance parameters: expression trees.

    Returns
    -------
    ET
    """
    if inst_args is None:
        inst_args = []
    _check_object(operator, 'create_call', 'operator', suite.Operator)
    args = _normalize_tree_ex(args) if args else []
    inst_args = _normalize_tree_ex(inst_args) if inst_args else []
    return _Operator(operator, args, inst_args, [])


def create_higher_order_call(
    operator: suite.Operator, args: LX, modifiers: Union[ET, List[ET]], inst_args: LX = None
) -> ET:
    """
    Return the expression tree for a call to an operator.

    Parameters
    ----------
    op : suite.Operator
        Operator to call.
    args : Union[EX, List[EX]]
        Parameters: expression trees.
    modifiers : Union[ET, List[ET]]
        Higher-order constructs: expression trees.
    inst_args : Union[EX, List[EX]]
        Instance parameters: expression trees.

    Returns
    -------
    ET
    """
    if inst_args is None:
        inst_args = []
    _check_object(operator, 'create_higher_order_call', 'operator', suite.Operator)
    args = _normalize_tree_ex(args) if args else []
    inst_args = _normalize_tree_ex(inst_args) if inst_args else []
    modifiers = modifiers if isinstance(modifiers, list) else [modifiers] if modifiers else []
    return _Operator(operator, args, inst_args, modifiers)


# arithmetic and logic


def create_unary(op: str, tree: EX, modifiers: Union[ET, List[ET]] = None) -> ET:
    """
    Return the expression tree for a unary operator.

    Parameters
    ----------
    op : str
        Unary operator to call: - | + | !
    tree : EX
        Operand: expression tree.
    modifiers : Union[ET, LIST[ET]], default: None
        List of higher-order constructs.

    Returns
    -------
    ET
    """
    eck = _unary_ops.get(op)
    if eck is None:
        raise ExprSyntaxError('create_unary', op)
    tree = _normalize_tree(tree)
    modifiers = modifiers if isinstance(modifiers, list) else [modifiers] if modifiers else []
    return _Predefined(eck, [tree], [], modifiers)


def create_binary(op: str, tree1: EX, tree2: EX, modifiers: Union[ET, List[ET]] = None) -> ET:
    """
    Return the expression tree for a binary operator.

    Parameters
    ----------
    op : str
        Binary operator to call: & | | | ^ | # | + | - | * | / | : | % | < | <= | > | >= | = | <>
    tree1 : EX
        First operand: expression tree.
    tree2 : EX
        Second operand: expression tree.
    modifiers : Union[ET, List[ET]], default: None
        List of higher-order constructs.

    Returns
    -------
    ET
    """
    eck = _binary_ops.get(op)
    if eck is None:
        raise ExprSyntaxError('create_binary', op)
    tree1, tree2 = _normalize_tree((tree1, tree2))
    modifiers = modifiers if isinstance(modifiers, list) else [modifiers] if modifiers else []
    return _Predefined(eck, [tree1, tree2], [], modifiers)


def create_nary(op: str, *args: List[EX], modifiers: Union[ET, List[ET]] = None) -> ET:
    """
    Return the expression tree for a nary operator.

    Parameters
    ----------
    op : str
        Nary operator to call: & | | | ^ | # | + | *
    args : List[EX]
        Operands: expression trees.
    modifiers : Union[ET, List[ET]], default: None
        List of higher-order constructs, to be provided as keyword parameter.

    Returns
    -------
    ET
    """
    eck = _nary_ops.get(op)
    if eck is None:
        raise ExprSyntaxError('create_nary', op)
    if len(args) < 2:
        raise ExprSyntaxError('create_nary', args)
    args = _normalize_tree(args)
    modifiers = modifiers if isinstance(modifiers, list) else [modifiers] if modifiers else []
    return _Predefined(eck, args, [], modifiers)


# selectors


def create_if(condition: EX, then: LX, else_: LX) -> ET:
    r"""
    Return the expression tree for the if-then-else operator.

    Notes
    -----
    This is an interface change with respect to the *SCADE Creation Library*.
    The ``then`` flows and ``else`` flows are now specified in two separate lists.

    Parameters
    ----------
    condition : EX
        Expression tree corresponding to the condition of the selector.
    then : Union[EX, List[EX]]
        List of expressions trees when the condition is ``True``.
    else\_ : Union[EX, List[EX]]
        List of expressions trees when the condition is ``False``.

    Returns
    -------
    ET
    """
    norm_then = _normalize_tree_ex(then)
    norm_else = _normalize_tree_ex(else_)
    length = len(norm_then)
    if length == 0 or length != len(norm_else):
        raise ExprSyntaxError('create_if', then)

    condition = _normalize_tree(condition)
    then_tree = _create_sequence(norm_then)
    else_tree = _create_sequence(norm_else)

    return _Predefined(Eck.IF, [condition, then_tree, else_tree], [], [])


def create_case(selector: EX, cases: List[Tuple[EX, EX]], default: EX = None) -> ET:
    """
    Return the expression tree for the case operator.

    Notes
    -----
    This is an interface change with respect to the *SCADE Creation Library*.
    The pairs "pattern"/"value" are now embedded in a list of tuples, and the
    default value is optional.

    Parameters
    ----------
    selector : EX
        Expression tree corresponding to the selector.
    cases : List[Tuple[EX, EX]]
        Pattern/values expression trees.
    default: EX, default:None
        Default value.

    Returns
    -------
    ET
    """
    if len(cases) == 0:
        raise ExprSyntaxError('create_case', cases)

    selector = _normalize_tree(selector)
    patterns = []
    inputs = []
    for pattern, input in cases:
        patterns.append(_normalize_tree(pattern))
        inputs.append(_normalize_tree(input))
    if default:
        patterns.append(_normalize_tree('_'))
        inputs.append(_normalize_tree(default))
    pattern_tree = _create_sequence(patterns)
    input_tree = _create_sequence(inputs)

    return _Predefined(Eck.CASE, [selector, input_tree, pattern_tree], [], [])


# types


def create_make(
    type_: suite.NamedType, *args: List[EX], modifiers: Union[ET, List[ET]] = None
) -> ET:
    r"""
    Return the expression tree for making a structured value.

    Parameters
    ----------
    type\_ : suite.NamedType
        Type to instantiate.
    args : List[EX]
        Values of the type instance.
    modifiers : Union[ET, List[ET]], default: None
        List of higher-order constructs, which is provided as a keyword parameter.

    Returns
    -------
    ET
    """
    if len(args) < 1:
        raise ExprSyntaxError('create_make', args)
    _check_object(type_, 'create_make', 'type', suite.NamedType)
    args, type_ = _normalize_tree((args, type_))
    modifiers = modifiers if isinstance(modifiers, list) else [modifiers] if modifiers else []
    return _Predefined(Eck.MAKE, [_create_sequence(args), type_], [], modifiers)


def create_flatten(type_: suite.NamedType, arg: EX, modifiers: Union[ET, List[ET]] = None) -> ET:
    r"""
    Return the expression tree for flattening a structured value.

    Parameters
    ----------
    type\_ : suite.NamedType
        Type to instantiate.
    arg : EX
        Value to flatten.
    modifiers : Union[ET, List[ET]], default: None
        List of higher-order constructs.

    Returns
    -------
    ET
    """
    _check_object(type_, 'create_flatten', 'type', suite.NamedType)
    arg, type_ = _normalize_tree((arg, type_))
    modifiers = modifiers if isinstance(modifiers, list) else [modifiers] if modifiers else []
    return _Predefined(Eck.FLATTEN, [arg, type_], [], modifiers)


# structures


def create_scalar_to_vector(size: EX, *args: List[EX]) -> ET:
    """
    Return the expression tree for the scalar-to-vector operator.

    Notes
    -----
    This is an interface change with respect to the *SCADE Creation Library*.
    The parameter size has moved from the last position to the first one.

    Parameters
    ----------
    size: EX
        Size of the vector.
    args : List[EX]
        Input values.

    Returns
    -------
    ET
    """
    if len(args) == 0:
        raise ExprSyntaxError('create_scalar_to_vector', args)
    size = _normalize_tree(size)
    args = _normalize_tree(args)
    # the size must be the last parameter
    return _Predefined(Eck.SCALAR_TO_VECTOR, list(args) + [size], [], [])


def create_data_array(*args: List[EX]) -> ET:
    """
    Return the expression tree for the data array operator.

    Parameters
    ----------
    args : List[EX]
        Values of the array.

    Returns
    -------
    ET
    """
    if len(args) == 0:
        raise ExprSyntaxError('create_data_array', args)
    args = _normalize_tree(args)
    return _Predefined(Eck.BLD_VECTOR, args, [], [])


def create_data_struct(*args: List[Tuple[str, EX]]) -> ET:
    """
    Return the expression tree for the data strictire operator.

    Notes
    -----
    This is an interface change with respect to the *SCADE Creation Library*.
    The pairs "name"/"value" are now embedded in a list of tuples.

    Parameters
    ----------
    args : List[Tuple[str, EX]]
        Label/values expression trees.

    Returns
    -------
    ET
    """
    length = len(args)
    if length == 0:
        raise ExprSyntaxError('create_data_struct', args)

    parameters = []
    for label, value in args:
        if not isinstance(label, str) or not label.isidentifier():
            raise TypeIdentifierError('create_data_struct', args)
        parameter = _normalize_tree(value)
        parameter.label = label
        parameters.append(parameter)
    return _Predefined(Eck.BLD_STRUCT, parameters, [], [])


def create_prj(flow: EX, path: LX) -> ET:
    """
    Return the expression tree for the projection operator.

    Parameters
    ----------
    flow : EX
        Input flow of the projection.
    path : Union[EX, List[EX]]
        Elements of the path, which is either the label or index.

    Returns
    -------
    ET
    """
    flow = _normalize_tree(flow)
    path = _normalize_tree_ex(path)
    parameters = [flow] + path
    return _Predefined(Eck.PRJ, parameters, [], [])


def create_prj_dyn(flow: EX, path: LX, default: EX) -> ET:
    """
    Return the expression tree for the dynamic projection operator.

    Parameters
    ----------
    flow : EX
        Input flow of the projection.
    path : Union[EX, List[EX]]
        Elements of the path, which is a label, index, or variable.
    default : EX
        Default value for the projection when the path is incorrect.

    Returns
    -------
    ET
    """
    flow, default = _normalize_tree((flow, default))
    path = _normalize_tree_ex(path)
    parameters = [flow] + path + [default]
    return _Predefined(Eck.PRJ_DYN, parameters, [], [])


def create_change_ith(flow: EX, path: LX, value: EX) -> ET:
    """
    Return the expression tree for the with operator.

    Parameters
    ----------
    flow : EX
        Input flow of the projection.
    path : Union[EX, List[EX]]
        Elements of the path, which is a label, index, or variable.
    value : EX
        Value to assign.

    Returns
    -------
    ET
    """
    flow, value = _normalize_tree((flow, value))
    path = _normalize_tree_ex(path)
    parameters = [flow, value] + path
    return _Predefined(Eck.CHANGE_ITH, parameters, [], [])


# time


def create_pre(*args: List[EX]) -> ET:
    """
    Return the expression tree for the pre operator.

    Parameters
    ----------
    args : List[EX]
        Input flows.

    Returns
    -------
    ET
    """
    if len(args) == 0:
        raise ExprSyntaxError('create_pre', '')
    args = _normalize_tree(args)
    return _Predefined(Eck.PRE, args, [], [])


def create_init(flows: LX, inits: LX) -> ET:
    """
    Return the expression tree for the init operator.

    Notes
    -----
    This is an interface change with respect to the *SCADE Creation Library*.
    The flows and their initial values are now specified in two separate lists.

    Parameters
    ----------
    flows : Union[EX, List[EX]]
        Input flows.
    inits : Union[EX, List[EX]]
        Initial values.

    Returns
    -------
    ET
    """
    norm_flows = _normalize_tree_ex(flows)
    norm_inits = _normalize_tree_ex(inits)
    length = len(norm_flows)
    if length == 0 or length != len(norm_inits):
        raise ExprSyntaxError('create_init', flows)

    flows_tree = _create_sequence(norm_flows)
    inits_tree = _create_sequence(norm_inits)

    return _Predefined(Eck.FOLLOW, [flows_tree, inits_tree], [], [])


def create_fby(flows: LX, delay: EX, inits: LX) -> ET:
    """
    Return the expression tree for the init operator.

    Notes
    -----
    This is an interface change with respect to the *SCADE Creation Library*.
    The flows and their initial values are now specified in two separate lists.

    Parameters
    ----------
    flows : Union[EX, List[EX]]
        Input flows.
    delay : EX
        Delay of the operator.
    inits : Union[EX, List[EX]]
        Initial values.

    Returns
    -------
    ET
    """
    norm_flows = _normalize_tree_ex(flows)
    norm_inits = _normalize_tree_ex(inits)
    length = len(norm_flows)
    if length == 0 or length != len(norm_inits):
        raise ExprSyntaxError('create_fby', flows)

    delay = _normalize_tree(delay)
    parameters = norm_flows + [delay] + norm_inits
    return _Predefined(Eck.FBY, parameters, [], [])


def create_times(number: EX, flow: EX) -> ET:
    """
    Return the expression tree for the times operator.

    Parameters
    ----------
    number : EX
        Number of cycles.
    flow : EX
        Input flow.

    Returns
    -------
    ET
    """
    number, flow = _normalize_tree((number, flow))
    return _Predefined(Eck.TIMES, [number, flow], [], [])


# array


def create_slice(array: EX, start: EX, end: EX) -> ET:
    """
    Return the expression tree for the slice operator.

    Parameters
    ----------
    array : EX
        Input array.
    start : EX
        Start index of the slice.
    end : EX
        End index of the slice.

    Returns
    -------
    ET
    """
    array, start, end = _normalize_tree((array, start, end))
    return _Predefined(Eck.SLICE, [array, start, end], [], [])


def create_concat(*args: List[EX]) -> ET:
    """
    Return the expression tree for the concat operator.

    Parameters
    ----------
    args : List[EX]
        Input arrays to concatenate.

    Returns
    -------
    ET
    """
    if len(args) < 2:
        raise ExprSyntaxError('create_concat', args)
    args = _normalize_tree(args)
    return _Predefined(Eck.CONCAT, args, [], [])


def create_reverse(flow: EX) -> ET:
    """
    Return the expression tree for the reverse operator.

    Parameters
    ----------
    flow : EX
        Input flow.

    Returns
    -------
    ET
    """
    flow = _normalize_tree(flow)
    return _Predefined(Eck.REVERSE, [flow], [], [])


def create_transpose(array: EX, dim1: EX, dim2: EX) -> ET:
    """
    Return the expression tree for the transpose operator.

    Parameters
    ----------
    array : EX
        Input array.
    dim1 : EX
        First dimension.
    dim2 : EX
        Second dimension.

    Returns
    -------
    ET
    """
    array, dim1, dim2 = _normalize_tree((array, dim1, dim2))
    return _Predefined(Eck.TRANSPOSE, [array, dim1, dim2], [], [])


# activation


def create_restart(every: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for restarting.

    Parameters
    ----------
    every : EX
        Input condition.

    Returns
    -------
    ET
    """
    every = _normalize_tree(every)
    return _Predefined(Eck.RESTART, [every], [], [])


def create_activate(every: EX, *args: List[EX]) -> ET:
    """
    Return the expression tree for the higher-order construct for activating with initial values.

    Parameters
    ----------
    every : EX
        Input condition.
    args: List[EX]
        Initial values.

    Returns
    -------
    ET
    """
    every = _normalize_tree(every)
    # args may be empty
    args = _normalize_tree(args) if args else args
    return _Predefined(Eck.ACTIVATE, [every, _create_sequence(args)], [], [])


def create_activate_no_init(every: EX, *args: List[EX]) -> ET:
    """
    Return the expression tree for the higher-order construct for activating with default values.

    Parameters
    ----------
    every : EX
        Input condition.
    args: List[EX]
        Default values.

    Returns
    -------
    ET
    """
    every = _normalize_tree(every)
    # args may be empty
    args = _normalize_tree(args) if args else args
    return _Predefined(Eck.ACTIVATE_NOINIT, [every, _create_sequence(args)], [], [])


# iterators


def create_map(size: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for map creation.

    Parameters
    ----------
    size : EX
        Number of iterations.

    Returns
    -------
    ET
    """
    size = _normalize_tree(size)
    return _Predefined(Eck.MAP, [size], [], [])


def create_mapi(size: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for mapi creation.

    Parameters
    ----------
    size : EX
        Number of iterations.

    Returns
    -------
    ET
    """
    size = _normalize_tree(size)
    return _Predefined(Eck.MAPI, [size], [], [])


def create_fold(size: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for fold creation.

    Parameters
    ----------
    size : EX
        Number of iterations.

    Returns
    -------
    ET
    """
    size = _normalize_tree(size)
    return _Predefined(Eck.FOLD, [size], [], [])


def create_foldi(size: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for foldi creation.

    Parameters
    ----------
    size : EX
        Number of iterations.

    Returns
    -------
    ET
    """
    size = _normalize_tree(size)
    return _Predefined(Eck.FOLDI, [size], [], [])


def create_mapfold(size: EX, acc: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for mapfold creation.

    Parameters
    ----------
    size : EX
        Number of iterations.
    acc : EX
        Number of accumulators.

    Returns
    -------
    ET
    """
    size, acc = _normalize_tree((size, acc))
    return _Predefined(Eck.MAPFOLD, [size, acc], [], [])


def create_mapfoldi(size: EX, acc: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for mapfoldi creation.

    Parameters
    ----------
    size : EX
        Number of iterations.

    acc : EX
        Number of accumulators.

    Returns
    -------
    ET
    """
    size, acc = _normalize_tree((size, acc))
    return _Predefined(Eck.MAPFOLDI, [size, acc], [], [])


def create_foldw(size: EX, condition: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for foldw creation.

    Parameters
    ----------
    size : EX
        Number of iterations.
    condition : EX
        Initial value of the iteration condition.

    Returns
    -------
    ET
    """
    size, condition = _normalize_tree((size, condition))
    return _Predefined(Eck.FOLDW, [size, condition], [], [])


def create_foldwi(size: EX, condition: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for foldwi creation.

    Parameters
    ----------
    size : EX
        Number of iterations.
    condition : EX
        Initial value of the iteration condition.

    Returns
    -------
    ET
    """
    size, condition = _normalize_tree((size, condition))
    return _Predefined(Eck.FOLDWI, [size, condition], [], [])


def create_mapw(size: EX, condition: EX, default: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for mapw creation.

    Parameters
    ----------
    size : EX
        Number of iterations.
    condition : EX
        Initial value of the iteration condition.
    default : EX
        Default value of the iteration.

    Returns
    -------
    ET
    """
    size, condition, default = _normalize_tree((size, condition, default))
    return _Predefined(Eck.MAPW, [size, condition, default], [], [])


def create_mapwi(size: EX, condition: EX, default: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for mapdwi creation.

    Parameters
    ----------
    size : EX
        Number of iterations.
    condition : EX
        Initial value of the iteration condition.
    default : EX
        Default value of the iteration.

    Returns
    -------
    ET
    """
    size, condition, default = _normalize_tree((size, condition, default))
    return _Predefined(Eck.MAPWI, [size, condition, default], [], [])


def create_mapfoldw(size: EX, acc: EX, condition: EX, default: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for mapfoldw creation.

    Parameters
    ----------
    size : EX
        Number of iterations.
    acc : EX
        Number of accumulators.
    condition : EX
        Initial value of the iteration condition.
    default : EX
        Default value of the iteration.

    Returns
    -------
    ET
    """
    size, acc, condition, default = _normalize_tree((size, acc, condition, default))
    return _Predefined(Eck.MAPFOLDW, [size, acc, condition, default], [], [])


def create_mapfoldwi(size: EX, acc: EX, condition: EX, default: EX) -> ET:
    """
    Return the expression tree for the higher-order construct for mapfoldwi creation.

    Parameters
    ----------
    size : EX
        Number of iterations.
    acc : EX
        Number of accumulators.
    condition : EX
        Initial value of the iteration condition.
    default : EX
        Default value of the iteration.

    Returns
    -------
    ET
    """
    size, acc, condition, default = _normalize_tree((size, acc, condition, default))
    return _Predefined(Eck.MAPFOLDWI, [size, acc, condition, default], [], [])


# ----------------------------------------------------------------------------
# Helpers (private)


def _create_sequence(flows: List[EX]) -> ET:
    """Create an expression tree for a group of flows."""
    return _Predefined(Eck.SEQ_EXPR, flows, [], [])


class ExprSyntaxError(Exception):
    """Provides the generic exception for syntax errors in expression trees."""

    def __init__(self, context, item):
        """Provide a customized message."""
        super().__init__('%s: %s: Syntax error' % (context, item))


class TypeIdentifierError(Exception):
    """Provides the exception for incorrect identifiers."""

    def __init__(self, context, item):
        """Provide a customized message."""
        super().__init__('%s: %s: Not a valid identifier' % (context, item))


class EmptyTreeError(Exception):
    """Provides the exception for empty expression trees."""

    def __init__(self):
        """Provide a customized message."""
        super().__init__('_normalize_tree: Illegal empty tree')


def _is_int(number: str) -> bool:
    # trivial test first
    try:
        _ = int(number)
        return True
    except ValueError:
        pass
    tokens = number.split('_', 1)
    if len(tokens) != 2:
        return False
    if tokens[1] in {'i8', 'i16', 'i32', 'i64', 'ui8', 'ui16', 'ui32', 'ui64'}:
        try:
            _ = int(tokens[0])
            return True
        except ValueError:
            pass
    return False


def _is_real(number: str) -> bool:
    # trivial test first
    try:
        _ = float(number)
        return True
    except ValueError:
        pass
    tokens = number.split('_', 1)
    if len(tokens) == 2 and tokens[1] in ('f32', 'f64'):
        try:
            _ = float(tokens[0])
            return True
        except ValueError:
            pass
    return False


def _is_bool(value: str) -> bool:
    # trivial test first
    return value == 'true' or value == 'false'


# TODO: modifiers
# TODO: move to query
def _find_expr_id(expr: suite.Expression, index: int) -> suite.ExprId:
    """
    Return the ``ExprId`` instance corresponding to the pin index of an equation.

    Notes
    -----
    This function is not currently complete and must be specified more precisely.

    Parameters
    ----------
    expr : suite.Expression
        Expression to consider, either an ExprId or ExprCall.
    index : int
        Pin number.

    Returns
    -------
    suite.ExprId
    """
    if isinstance(expr, suite.ExprId):
        # the expression itself!
        # index shall be 0
        return expr if index == 0 else None
    elif not isinstance(expr, suite.ExprCall):
        # shall not occur: return the expression
        # the caller is responsible for displaying the error
        return None

    # expr is an ExprCall
    params = expr.parameters
    code = Eck(expr.predef_opr)
    if code == Eck.IF:
        expr_cond, expr_then, expr_else = params
        # rebuild a new list of parameters
        params = [expr_cond]
        params.extend(expr_then.parameters)
        params.extend(expr_else.parameters)
    elif code == Eck.CASE:
        expr_selector, expr_inputs, _ = params
        # rebuild a new list of parameters
        params = [expr_selector]
        params.extend(expr_inputs.parameters)
    elif code == Eck.MAKE:
        expr_inputs, _ = params
        # rebuild a new list of parameters
        params = expr_inputs.parameters
    elif code == Eck.BLD_STRUCT:
        # rebuild a new list of parameters
        params = (value for name, value in params)
    elif code == Eck.FOLLOW:
        expr_flows, expr_inits = params
        # rebuild a new list of parameters
        params = expr_flows.parameters + expr_inits.parameters
    else:
        # operator call
        # unary, n-ary...
        # "FLATTEN" - "SCALAR_TO_VECTOR" - "BLD_VECTOR" - "PRJ" - "PRJ_DYN" - "CHANGE_ITH" -
        # "PRE" - "FBY" - "TIMES" - "SLICE" - "CONCAT" - "REVERSE" - "TRANSPOSE"
        # (includes operator call)
        # ith parameter (TODO: check index vs [llength $params])
        # list params left unchanged
        pass
    # TODO RESTART, ACTIVATE, ACTIVATE_NO_INIT, MAP/FOLD*
    return params[index]



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides numeric values for the ``predef_opr`` attribute from the ``suite.ExprCall``.

For more information, see the SCADE Suite documentation:
``5 Specific Commands for Python Scripting > Access to Predefined Operators in Python Scripts.``
"""

from enum import Enum


class Eck(Enum):
    """Provides an enum of predefined operators."""

    NONE = 1
    AND = 2
    OR = 3
    XOR = 4
    NOT = 5
    SHARP = 6
    PLUS = 7
    SUB = 8
    NEG = 9
    MUL = 10
    REAL2INT = 11
    INT2REAL = 12
    SLASH = 14
    DIV = 15
    MOD = 16
    PRJ = 18
    CHANGE_ITH = 19
    LESS = 20
    LEQUAL = 21
    GREAT = 22
    GEQUAL = 23
    EQUAL = 24
    NEQUAL = 25
    PRE = 26
    WHEN = 28
    FOLLOW = 29
    FBY = 30
    IF = 31
    CASE = 32
    SEQ_EXPR = 33
    BLD_STRUCT = 34
    MAP = 35
    FOLD = 36
    MAPFOLD = 37
    MAPI = 38
    FOLDI = 39
    SCALAR_TO_VECTOR = 40
    BLD_VECTOR = 41
    PRJ_DYN = 42
    MAKE = 43
    FLATTEN = 44
    MERGE = 45
    REVERSE = 46
    TRANSPOSE = 47
    TIMES = 49
    MATCH = 50
    SLICE = 51
    CONCAT = 52
    ACTIVATE = 53
    RESTART = 54
    FOLDW = 55
    FOLDWI = 56
    ACTIVATE_NOINIT = 57
    CLOCKED_ACTIVATE = 58
    CLOCKED_NOT = 59
    POS = 60
    MAPW = 61
    MAPWI = 62
    NUMERIC_CAST = 63
    MAPFOLDI = 64
    MAPFOLDW = 65
    MAPFOLDWI = 66
    LAND = 67
    LOR = 68
    LXOR = 69
    LNOT = 70
    LSL = 71
    LSR = 72



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides helpers for SCADE model creation functions.

* Persistence
* Predefined types
* Type trees
"""

from enum import Enum
from os.path import abspath, relpath
from pathlib import Path
from typing import Union

import scade.model.project.stdproject as std
import scade.model.suite as suite

from .. import _scade_api
from .project import _check_object, _find_file_ref, create_file_ref, create_folder

_modified_files = set()
"""
Set of modified fimes for save_all.

The set is reset once the files are saved.
"""

_pending_links = []
"""
List of pending associations to be set once an object is built.

* Each element of the list is a tuple (src: Object, role: str, link: Object).
* The list is reset by the ``link_pendings`` function.
"""


# ----------------------------------------------------------------------------
# Interface


class Sfk(Enum):
    """Provides kinds of simulation files."""

    C = 1  # Source File for C
    OBJECT = 2  # Object File
    MACRO = 3  # Macro File
    TYPE = 4  # Type Definition File
    ADA = 5  # Source File for Ada


def save_all():
    """Save all modified files of Scade models."""
    global _modified_files

    for unit in _modified_files:
        unit.save()

    _modified_files = set()


def add_element_to_project(
    project: std.Project,
    element: suite.StorageElement,
    folder: std.Folder = None,
    default: bool = True,
) -> std.FileRef:
    """
    Add the file defining the storage element to the project.

    The file is added if and only if it is not already present.
    If the file is present, the ``folder`` and ``default`` parameters
    are ignored.

    * When the ``folder`` parameter is not ``None``, the file is added to the
      folder specified.
    * When the ``default`` parameter is ``True``, the file is added to either the
      "Model Files" or "Separate Files" folder, depending on the nature of the
      storage unit.
    * Otherwise, the file is added to the project as a root element.

    Parameters
    ----------
    project : std.Project
        Project to modify.
    element : suite.StorageELement
        Element with storage unit to add to the project.
        If the storage unit is not a root element, the added file is
        tagged as ``NONROOT`` for SCADE.
    folder : std.Folder, default: None
        Parent folder of the file to add to the project.
    default : bool, default: True
        Whether to add the file to one of the default folders
        for SCADE Suite files.

    Returns
    -------
    std.FileRef
    """
    unit = element.defined_in
    path = unit.sao_file_name
    file_ref = _find_file_ref(project, Path(path))
    if not file_ref:
        if folder:
            parent = folder
        elif default:
            names = ['Model Files']
            if not unit.is_root():
                names.append('Separate Files')
            parent = create_folder(project, names)
        else:
            parent = project
        file_ref = create_file_ref(parent, path)
        # update persist_as to a relative path to the project
        file_ref.set_path_name(path)
        if not unit.is_root():
            file_ref.set_bool_tool_prop_def('SCADE', 'NONROOT', True, False, None)
    return file_ref


def add_imported_to_project(
    project: std.Project,
    element: Union[suite.NamedType, suite.Operator],
    path: str,
    folder: std.Folder = None,
    default: bool = True,
) -> std.FileRef:
    """
    Add an imported source file, associated with the SCADE imported element, to the project.

    The file is added if and only if it is not already present.
    If the file is present, the ``folder`` and ``default`` parameters are ignored.

    * When the ``folder`` parameter is not ``None``, the file is added to the
      folder specified.
    * When the ``default`` parameter is ``True``, the file is added to either the
      "External Code" or "External Type Definitions" folder, depending on the nature
      of the element.
    * Otherwise, the file is added to the project as a root element.

    Parameters
    ----------
    project : std.Project
        Project to modify.
    element : Union[suite.NamedType, suite.Operator]
        Imported element.
    path : str
        Path of the file to add to the project.
    folder : std.Folder
        Parent folder of the file to add to the project.
    default : bool, default: True
        Whether to add the file is added to the default folder
        for SCADE Simulation files, according to the element.

    Returns
    -------
    std.FileRef
    """
    if isinstance(element, suite.Operator):
        kind = Sfk.ADA if Path(path).suffix.lower() in ['.adb', '.ads'] else Sfk.C
        prop = 'IMPORTED_NODES'
    else:
        kind = Sfk.TYPE
        prop = 'IMPORTED_TYPES'

    file_ref = add_simulation_file_to_project(project, path, kind, folder, default)

    elements = file_ref.get_tool_prop_def('SCADE', prop, [], None)
    # get the path without the trailing slash, so that the IDE
    # does fail to find the relationship file-element
    elements.append(element.get_full_path().strip('/'))
    file_ref.set_tool_prop_def('SCADE', prop, elements, [], None)

    return file_ref


def add_simulation_file_to_project(
    project: std.Project, path: str, kind: Sfk, folder: std.Folder = None, default=True
) -> std.FileRef:
    """
    Add a file to the project and tag it appropriately for the SCADE simulation.

    The file is added if and only if it is not already present.
    If it is present, the parameters folder and default are ignored.

    The file is added to:

    * The specified folder when not None.
    * Or one of the default folders  "External Code" or "External Type Definitions",
      depending on the kind of the file.
    * Otherwise to the project as a root element.

    Parameters
    ----------
    project : std.Project
        Project to modify.
    path : Path
        Path of the file to be added to the project.
    kind: Sfk
        Kind of the added file.
    folder : std.Folder
        Parent folder of the file to add to the project.
    default : bool
        When True, the file is added to the default folder
        for SCADE Simulation files, according to kind.

    Returns
    -------
    std.FileRef
    """
    file_ref = _find_file_ref(project, path)
    if not file_ref:
        if folder:
            parent = folder
        elif default:
            folders = {
                Sfk.C: 'External Code',
                Sfk.OBJECT: 'External Type Code',
                Sfk.MACRO: 'External Definitions',
                Sfk.TYPE: 'External Type Definitions',
                Sfk.ADA: 'External Code',
            }
            parent = create_folder(project, folders[kind])
        else:
            parent = project
        file_ref = create_file_ref(parent, path)
        # update persist_as to a relative path to the project
        file_ref.set_path_name(path)
        values = {
            Sfk.C: 'CS',
            Sfk.OBJECT: 'Obj',
            Sfk.MACRO: 'Macro',
            Sfk.TYPE: 'Type',
            Sfk.ADA: 'AdaS',
        }
        file_ref.set_scalar_tool_prop_def('SIMULATOR', 'FILEKIND', values[kind], '', None)

    return file_ref


# ----------------------------------------------------------------------------
# Helpers (private)


def _get_owner_model(object_: suite.Object) -> suite.Model:
    r"""
    Get the model owning an object.

    Parameters
    ----------
    object\_ : suite.Object
        Input object.

    Returns
    -------
    suite.Model
    """
    # obvious...
    _check_object(object_, '_get_owner_model', 'obejct_', suite.Object)
    if isinstance(object_, suite.Model):
        return object_
    # shortcut
    if object_.defined_in:
        return object_.defined_in.model
    # the item is not associated to a file, for example a predefined type?
    # --> loop on the owners
    owner = object_.owner
    while owner and not isinstance(owner, suite.Model):
        owner = owner.owner
    return owner


def _get_model_project(model: suite.Model) -> std.Project:
    """
    Get the project adssociated with a model or ``None`` if the model is a library.

    Notes
    -----
    This function is an alternative to ``model.project``, which causes a crash
    in some circumstances.

    Parameters
    ----------
    model : suite.Model
        Input model.

    Returns
    -------
    std.Project
    """
    pathname = abspath(model.descriptor.model_file_name)
    return next((_ for _ in std.get_roots() if abspath(_.pathname) == pathname), None)


def _get_default_file(model: suite.Model) -> Path:
    """
    Get the default file to store model-level declarations.

    Notes
    -----
    This function returns either the value of the projet's tool property, ``@SCADE:DEFAULTFILE``,
    or the path of the model with the suffix ``.xscade``.

    Parameters
    ----------
    model : suite.Model
        Input model.

    Returns
    -------
    str
    """
    # default value
    path = Path(model.descriptor.model_file_name).with_suffix('.xscade')
    # check for an overridden value in the projet's properties
    # model.project may crash in some circumstances
    project = _get_model_project(model)
    if project:
        value = project.get_scalar_tool_prop_def('SCADE', 'DEFAULTFILE', path.name, None)
        # make the path absolute
        path = path.parent / value

    return path


def _link_storage_element(owner: suite.Package, element: suite.StorageElement, path: Path):
    """
    Link a storage element to the storage unit specified by a path.

    * The owner is either a package or a model.
    * The storage unit is created if it does not exist

    Notes
    -----
    The element can by any model declaration if the owner is a model.
    Otherwise, the element must be a package or an operator.

    Parameters
    ----------
    owner : suite.Package
        Owner of the storage element.
    element : suite.StorageELement
        Element to add to the storage unit.
    path : Path
        Path of the storage unit.

    Returns
    -------
    str
    """
    global _modified_files

    # get the parent model (must exist)
    model = _get_owner_model(owner)

    if owner == model and path is None:
        path = _get_default_file(model)

    if path is not None:
        if owner != model:
            directory = Path(owner.defined_in.sao_file_name).parent
            persist_as = str(Path(relpath(abspath(path), directory)).with_suffix(''))
        else:
            persist_as = ''
        unit = _create_unit(model, str(path), persist_as)
        element.storage_unit = unit
        _modified_files.add(unit)
    else:
        assert owner != model
    if owner != model:
        _set_modified(owner)


def _add_pending_link(object_: suite.Object, role: str, link: suite.Object):
    r"""
    Buffer the elements of an association.

    This function is used while creating some complex data. Each link from data elements
    is created once the complex data is built and connected to its owner.

    Indeed, if an error occurs while creating the data, it can be difficult to
    delete links to already exiting data, such as types.

    Parameters
    ----------
    object\_ : suite.Object
        Source of the association.
    role : str
        Name of the association end.
    link : suite.Object
        Destination of the association.
    """
    global _pending_links

    _pending_links.append((object_, role, link))


def _link_pendings():
    """Flush the pending links buffer."""
    global _pending_links

    for object_, role, link in _pending_links:
        _scade_api.set(object_, role, link)
    _pending_links = []


def _set_modified(object_: suite.Object):
    r"""
    Tag the file defining the input object as modified.

    Parameters
    ----------
    object\_ : suite.Object
        Input object.
    """
    global _modified_files

    unit = object_.defined_in
    if unit:
        unit.sao_modified = True
        _modified_files.add(unit)


def _create_unit(model: suite.Model, path: Path, persist_as: str) -> suite.StorageUnit:
    """
    Add a storage unit to a Scade model if it does not exist.

    Parameters
    ----------
    model : suite.Model
        Scade model.
    path : Path
        Path of the storage unit.
    persist_as: str
        Reference of the file to store in its owner file.

    Returns
    -------
    suite.StorageUnit
    """
    _check_object(model, '_create_unit', 'model', suite.Model)

    global _modified_files

    for unit in model.model_storage_units:
        if abspath(unit.sao_file_name) == abspath(path):
            _modified_files.add(unit)
            return unit
    unit = suite.StorageUnit(model)
    unit.sao_file_name = str(path)
    unit.persist_as = persist_as
    unit.model = model
    _modified_files.add(unit)

    return unit



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides ``scade_env.load_project``, renamed to ``declare_project``.

Search for a SCADE installation compatible with the Python version
and add it to ``sys.path`` if not already present.
Present means ``scade_env.pyd`` is accessible from ``sys.path``.
"""

import importlib.util
from pathlib import Path
import platform
import re
import sys
from typing import Optional

if platform.system() == 'Windows':
    import winreg as reg
else:
    # allow importing the file on other systems
    pass


def _resolve_venv(home: Path) -> Optional[Path]:
    """Get the virtual environment's target, if any."""
    cfg = home / 'pyvenv.cfg'
    if cfg.exists():
        for line in cfg.open('r'):
            m = re.match(r'^home\s*=\s*(.*)$', line)
            if m:
                return Path(m.groups()[0]).parent
    return None


def _get_scade_dirs(min='00.0', max='99.9'):
    """Get the list of SCADE installation directories."""
    names = []
    if platform.system() == 'Windows':
        for company in 'Esterel Technologies', 'Ansys Inc':
            try:
                hklm = reg.OpenKey(reg.HKEY_LOCAL_MACHINE, r'SOFTWARE\%s\SCADE' % company)
            except OSError:
                continue
            for i in range(reg.QueryInfoKey(hklm)[0]):
                name = reg.EnumKey(hklm, i)
                try:
                    dir, _ = reg.QueryValueEx(reg.OpenKey(hklm, name), 'InstallDir')
                    names.append((name, dir))
                except FileNotFoundError:
                    pass
    dirs = []
    for name, dir in sorted(names, key=lambda x: x[0]):
        if name >= min and name < max:
            dirs.append(dir)
            # print('short list', dir)
    return dirs


def _get_python_scade_versions(major: int, minor: int):
    """Get the highest SCADE installation compatible with a version of Python."""
    # [min, max[ releases for a given version of Python
    releases = {
        4: ('19.2', '21.2'),
        7: ('21.2', '23.2'),
        10: ('23.2', '26.1'),
        12: ('26.1', '99.9'),
    }
    if major != 3:
        return None
    if minor > 12:
        # starting 3.12, SCADE uses the Python Limited API
        minor = 12
    interval = releases.get(minor)
    return interval


def _get_compatible_scade_home(major: int, minor: int) -> Optional[Path]:
    """Get the most recent version of SCADE compatible with the input Python version."""
    interval = _get_python_scade_versions(major, minor)
    if interval:
        dirs = _get_scade_dirs(*interval)
        if dirs:
            # ordered list, get the most recent compatible version
            # and leave a message for who might read it
            # print('Using', dirs[-1])
            return Path(dirs[-1])
    return None


def _add_scade_to_sys_path():
    """Add to ``sys.path`` an installed SCADE release compatible with the current interpreter."""
    # resolve virtual environments to get SCADE's Python interpreter
    home = Path(sys.executable).parent.parent
    _wrapped_home = _resolve_venv(home)
    while _wrapped_home:
        home = _wrapped_home
        _wrapped_home = _resolve_venv(home)

    if home.name == 'contrib':
        # python.exe is located in <SCADE install>/contrib/Python3x
        home = home.parent
    else:
        # last chance, try the most recent installation of SCADE
        major, minor, _, _, _ = sys.version_info
        home = _get_compatible_scade_home(major, minor)

    if not home:  # pragma no cover
        # wrong installation or SCADE not available on the computer
        print('Use a Python interpreter delivered with SCADE.')
    else:
        # regular SCADE installation
        _base = home / 'SCADE'
        sys.path.append(str(_base / 'bin'))
        sys.path.append(str(_base / 'APIs' / 'Python' / 'lib'))


if platform.system() == 'Windows':
    # scade_env.pyd is in <scade installation>/SCADE/bin
    if not importlib.util.find_spec("scade_env"):
        _add_scade_to_sys_path()

    from scade_env import load_project as declare_project

    # import also _scade_api and scade, to avoid import order constraint in client files
    # isort: split
    import _scade_api
    import scade
else:
    # allow importing the file on other systems
    # for documentation generation, for example
    declare_project = None
    scade = None
    _scade_api = None



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a collection of functions for creating or updating Scade models."""

# ignore F401: functions made available for modules, not used here
from .data_def import (  # noqa: F401
    AK,
    DK,
    IT,
    SK,
    TD,
    TK,
    TR,
    IfTree,
    TransitionDestination,
    TransitionTree,
    add_data_def_assertion,
    add_data_def_equation,
    add_data_def_if_block,
    add_data_def_locals,
    add_data_def_net_diagram,
    add_data_def_probes,
    add_data_def_signals,
    add_data_def_state_machine,
    add_data_def_text_diagram,
    add_data_def_when_block,
    add_diagram_edge,
    add_diagram_equation_set,
    add_diagram_missing_edges,
    add_state_machine_state,
    add_state_transition,
    add_transition_equation,
    add_when_block_branches,
    create_if_action,
    create_if_tree,
    create_transition_fork,
    create_transition_state,
    create_when_branch,
    set_variable_default,
    set_variable_last,
)
from .declaration import (  # noqa: F401
    VK,
    IllegalIOError,
    ParamImportedError,
    add_enumeration_values,
    add_operator_hidden,
    add_operator_inputs,
    add_operator_outputs,
    add_operator_parameters,
    create_constant,
    create_enumeration,
    create_graphical_operator,
    create_imported_constant,
    create_imported_operator,
    create_imported_type,
    create_named_type,
    create_package,
    create_sensor,
    create_textual_operator,
    set_specialized_operator,
    set_type_constraint,
)
from .expression import (  # noqa: F401
    EmptyTreeError,
    ExprSyntaxError,
    TypeIdentifierError,
    create_activate,
    create_activate_no_init,
    create_binary,
    create_call,
    create_case,
    create_change_ith,
    create_concat,
    create_data_array,
    create_data_struct,
    create_fby,
    create_flatten,
    create_fold,
    create_foldi,
    create_foldw,
    create_foldwi,
    create_higher_order_call,
    create_if,
    create_init,
    create_make,
    create_map,
    create_mapfold,
    create_mapfoldi,
    create_mapfoldw,
    create_mapfoldwi,
    create_mapi,
    create_mapw,
    create_mapwi,
    create_nary,
    create_pre,
    create_prj,
    create_prj_dyn,
    create_restart,
    create_reverse,
    create_scalar_to_vector,
    create_slice,
    create_times,
    create_transpose,
    create_unary,
)
from .project import (  # noqa: F401
    create_configuration,
    create_file_ref,
    create_folder,
    create_prop,
    save_project,
)
from .scade import (  # noqa: F401
    add_element_to_project,
    add_imported_to_project,
    add_simulation_file_to_project,
    save_all,
)
from .type import create_sized, create_structure, create_table  # noqa: F401



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provide helpers for creating type trees.

Expression trees are intermediate structures to declare any arbitrary complex
types. They create the corresponding SCADE Suite type in the context of a
model element, such as the type of a constant.

This module provides functions to create a type tree for any type of the Scade
language. Thus, the intermediate structures or classes defining the type trees
can be opaque.

Notes: The typing is relaxed in this module to ease the constructs.

* ``TT`` is an alias for ``TypeTree`` to shorten the declarations.

* ``TX``, which stands for extended type tree, is defined as follows::

     Union[str, suite.Type, TT]

  This enhances the usability of these functions by accepting some values,
  such as existing types or name of predefined types, as valid type trees.

"""

from typing import List, Tuple, Union

import scade.model.suite as suite

from .expression import EX, _build_expression
from .scade import _add_pending_link, _get_owner_model


# type trees
class TypeTree:
    """Provides the top-level abstract class for type trees."""

    def _build_type(self, context: suite.Object) -> suite.Type:
        """Build a SCADE Suite type from the type tree."""
        # must be overridden
        assert False  # pragma no cover


TT = TypeTree
"""Short name for a ``TypeTree`` instance to simplify the declarations."""

TX = Union[str, suite.Type, TT]
"""Extended type tree to simply the use of the create functions."""


class _Type(TT):
    """Existing type."""

    def __init__(self, type_: suite.Type):
        """Literal value."""
        self.type = type_

    def _build_type(self, context: suite.Object) -> suite.Type:
        """Build a SCADE Suite type from the type tree."""
        return self.type


class _Predefined(TT):
    """Provides the predefined type."""

    def __init__(self, name: str):
        """Initialize a predefined type."""
        self.name = name

    def _build_type(self, context: suite.Object) -> suite.Type:
        """Build a SCADE Suite type from the type tree."""
        return _get_predefined_type(context, self.name)


class _Sized(TT):
    """Provides the sized type."""

    def __init__(self, signed: bool, size: EX):
        self.signed = signed
        self.size = size

    def _build_type(self, context: suite.Object) -> suite.Type:
        """Build a SCADE Suite type from the type tree."""
        type_ = suite.SizedType(context)
        type_.constraint = _get_type_constraint(context, 'signed' if self.signed else 'unsigned')
        type_.size_expression = _build_expression(self.size, context)

        return type_


class _Table(TT):
    """Provides the multi-dimensional array."""

    def __init__(self, dimensions: List[EX], type_: TT):
        self.dimensions = dimensions
        self.type = type_

    def _build_type(self, context: suite.Object) -> suite.Type:
        """Build a SCADE Suite type from the type tree."""
        subtype = self.type._build_type(context) if self.type else None
        for dim in self.dimensions:
            type_ = suite.Table(context)
            type_.size_expression = _build_expression(dim, context)
            _object_link_type(type_, subtype)
            subtype = type_

        return type_


class _Structure(TT):
    """Provides the structure."""

    def __init__(self, fields: List[Tuple[str, TT]]):
        self.fields = fields

    def _build_type(self, context: suite.Object) -> suite.Type:
        """Build a SCADE Suite type from the type tree."""
        type_ = suite.Structure(context)
        elements = []
        for name, subtree in self.fields:
            element = suite.CompositeElement(type_)
            element.name = name
            elements.append(element)
            subtype = subtree._build_type(context) if subtree else None
            _object_link_type(element, subtype)
        type_.elements = elements

        return type_


def _normalize_tree(any: TX) -> TT:
    """Create type tree instances from predefined types or SCADE objects."""
    if any is None or isinstance(any, TT):
        return any
    # SCADE objects
    if isinstance(any, suite.Type):
        return _Type(any)
    # predefined types
    if isinstance(any, str):
        if any == 'bool' or any == 'char':
            return _Predefined(any)
        elif any == 'int' or any == 'real':
            # KCG 6.4 and earlier
            return _Predefined(any)
        elif any in _numeric_types:
            # KCG 6.5 and greater
            return _Predefined(any)
        elif any and any[0] == "'":
            raise _polymorphic_error('_normalize_tree', any)

        else:
            raise _syntax_error('_normalize_tree', any)

    # fall through
    raise _syntax_error('_normalize_tree', any)


def _build_type(tree: TX, context: suite.Object) -> suite.Type:
    """
    Build a type from an extended type tree.

    Parameters
    ----------
    tree : TX
        Type expressed as an extended type tree.
    context : suite.Object
        Context of the creation of the type. This parameter is used to
        create the instances of the types, find the predefined types, or
        resolve polymorphic types (TODO).

    Returns
    -------
    suite.Type
    """
    tree = _normalize_tree(tree)
    return tree._build_type(context)


def create_sized(signed: bool, size: EX) -> TT:
    """
    Get the type tree for a sized type.

    Parameters
    ----------
    signed : bool
        Whether the type is signed.
    size : EX
        Size of the type expressed as an expression tree.

    Returns
    -------
    TT
    """
    if isinstance(size, int) and size not in [8, 16, 32, 64]:
        raise _syntax_error('_create_sized', size)
    return _Sized(signed, size)


def create_table(dimensions: Union[EX, List[EX]], type_: TX) -> TT:
    r"""
    Get the type tree for a structure.

    Parameters
    ----------
    type\_ : TX
        Type tree defining the type of the array elements.
    dimensions : Union[EX, List[EX]]
        Dimensions of the array, which is either a single expression tree or
        a list of expression trees.

    Returns
    -------
    TT
    """
    if not isinstance(dimensions, list):
        dimensions = [dimensions]
    if len(dimensions) == 0:
        raise _syntax_error('_create_table', dimensions)
    type_ = _normalize_tree(type_)
    return _Table(dimensions, type_)


def create_structure(*fields: List[Tuple[str, TX]]) -> TT:
    """
    Get the type tree for a structure.

    Notes
    -----
    This is an interface change with respect to the **SCADE Creation Library**.
    The pairs "pattern"/"value" are now embedded in a list of tuples.

    Parameters
    ----------
    fields : List[Tuple[str, TX]]
        Name/type expression trees.

    Returns
    -------
    TT
    """
    if len(fields) == 0:
        raise _syntax_error('create_structure', fields)
    fields = [(_[0], _normalize_tree(_[1])) for _ in fields]
    return _Structure(fields)


# ----------------------------------------------------------------------------
# Helpers (private)


def _syntax_error(context, tree) -> Exception:
    """Format a dedicated message from the parameters."""
    return ValueError('%s: %s: Type syntax error' % (context, tree))


def _polymorphic_error(context, tree) -> Exception:
    """Format a dedicated message from the parameters."""
    return ValueError('%s: %s: Illegal polymorphic type' % (context, tree))


_predefined_types = {}
"""
Cache for predefined types.

* The key is a name.
* The value is the corresponding predefined type.
"""

_type_constraints = {}
"""
Cache for type constraints.

* The key is a name.
* The value is the corresponding type constraint.
"""

_numeric_types = [
    'int8',
    'int16',
    'int32',
    'int64',
    'uint8',
    'uint16',
    'uint32',
    'uint64',
    'float32',
    'float64',
]
"""List of the numeric types for KCG 6.6."""

_constraints = [
    'numeric',
    'integer',
    'signed',
    'unsigned',
    'float',
]
"""List of constraints."""


def _get_predefined_type(context: suite.Object, name: str) -> suite.NamedType:
    """
    Return a predefined type instance from a name for a given context.

    Parameters
    ----------
    context : suite.Object
        Object from which the owning session can be derived.
    name : str
        Name of the type.

    Returns
    -------
    suite.NamedType
    """
    global _predefined_types

    type_ = _predefined_types.get(name)
    if type_ is None:
        model = _get_owner_model(context)
        session = model.session
        type_ = session.find_predefined_type(name)
        _predefined_types[name] = type_
    return type_


def _get_type_constraint(context: suite.Object, name: str) -> suite.TypeConstraint:
    """
    Get a type constraint instance from a name for a given context.

    Parameters
    ----------
    context : suite.Object
        Object from which the owning session can be derived.
    name : str
        Name of the type.

    Returns
    -------
    suite.TypeConstraint
    """
    global _type_constraints

    constraint = _type_constraints.get(name)
    if constraint is None:
        model = _get_owner_model(context)
        session = model.session
        constraint = session.find_type_constraint(name)
        _type_constraints[name] = constraint
    return constraint


def _object_link_type(object_: suite.TypedObject, type_: suite.Type):
    r"""
    Set the type of an object.

    * For an association object, the type is buffered.
    * For a composition object, the build type is updated immediately when required.

    Parameters
    ----------
    object\_ : suite.TypedObject
        Input object.
    type\_ : suite.Type
        Type of the object.
    """
    if type_:
        global _pending_links

        if not isinstance(type_, suite.NamedType):
            object_.build_type = type_
        _add_pending_link(object_, 'type', type_)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides functions for adding elements to a SCADE project (ETP file).

These functions do not check for semantic errors, like adding two files with the same path.
"""

from os.path import abspath
from typing import Any, List, Tuple, Union

import scade.model.project.stdproject as std


def create_folder(
    owner: Union[std.Project, std.Folder], path: Union[str, List[str]], extensions: str = ''
) -> std.Folder:
    """
    Create a folder.

    A folder has a name and must be either a root folder of a project or
    a subfolder of another folder.

    Parameters
    ----------
    owner : std.Project or std.Folder
        Owner of the folder, which is either the project itself or a folder.
    path : Union[str, List[str]]
        Path to or name of the folder. When a path is provided, the function
        creates the intermediate folders that do not exist.
    extensions: str
        String defining the extensions associated with the folder.

    Returns
    -------
    std.Folder
        Folder with the given path or name if this folder exists.
    """
    _check_object(owner, 'create_folder', 'owner', (std.Project, std.Folder))

    if isinstance(path, str):
        path = [path]
    for name in path[:-1]:
        # create the intermediate folders
        print('create intermediate', name)
        owner = create_folder(owner, name)
    name = path[-1]
    # check for the existence of the folder
    elements = _get_elements(owner)
    for element in elements:
        if isinstance(element, std.Folder) and element.name == name:
            return element

    # not found
    if isinstance(owner, std.Project):
        folder = std.Folder(owner)
        folder.owner = owner
    else:
        assert isinstance(owner, std.Folder)
        folder = std.Folder(owner.project)
        folder.folder = owner

    folder.name = name
    folder.extensions = extensions

    return folder


def create_file_ref(owner: Union[std.Project, std.Folder], persist_as: str) -> std.FileRef:
    """
    Create file reference.

    A file reference has a pathname and must be a file of either a project or a folder.

    Parameters
    ----------
    owner : std.Project or std.Folder
        Owner of the file, which is either the project itself or a folder.
    persist_as : str
        String representation of the reference to the file to store in the project's file.
        The string can be either a relative reference to the project or an absolute path.

    Returns
    -------
    std.FileRef
    """
    _check_object(owner, 'create_file_ref', 'owner', (std.Project, std.Folder))

    if isinstance(owner, std.Project):
        file_ref = std.FileRef(owner)
        file_ref.owner = owner
    else:
        assert isinstance(owner, std.Folder)
        file_ref = std.FileRef(owner.project)
        file_ref.folder = owner

    file_ref.persist_as = persist_as

    return file_ref


def create_configuration(owner: std.Project, name: str) -> std.Configuration:
    """
    Create a configuration.

    A configuration has a name and belongs to a project.

    Parameters
    ----------
    owner : std.Project
        Project.
    name : str
        Name of the configuration.

    Returns
    -------
    std.Configuration
    """
    _check_object(owner, 'create_configuration', 'owner', std.Project)

    configuration = std.Configuration(owner)
    configuration.project = owner
    configuration.name = name

    return configuration


def create_prop(
    owner: std.Annotable, configuration: std.Configuration, name: str, values: List[str]
) -> std.Prop:
    """
    Create a property.

    Properties are attached to a project, folder, or file reference.

    They have a name and a list of values. They can be associated with a configuration.

    Parameters
    ----------
    owner : std.Annotable
        Element to attach the property to.
    configuration : str
        Configuration to associate with the property or ``None``.
    name : str
        Name of the property.
    values : List[str]
        Values of the property as a list of strings.

    Returns
    -------
    std.Prop
    """
    _check_object(owner, 'create_prop', 'owner', std.Annotable)
    if configuration is not None:
        _check_object(configuration, 'create_prop', 'configuration', std.Configuration)

    project = owner if isinstance(owner, std.Project) else owner.project

    prop = std.Prop(project)
    prop.entity = owner
    prop.name = name
    prop.values = values

    if configuration is not None:
        prop.configuration = configuration

    return prop


def save_project(project: std.Project):
    """
    Save the project.

    The path of the project is specified in project.pathname.

    Parameters
    ----------
    project : std.Project
        Project to save.
    """
    _check_object(project, 'save_project', 'project', std.Project)
    project.save(project.pathname)


# -----------------------------------------------------------------------------
# Misc.


# not sure this is a good idea... keep this function for tests only?
def _create_empty_project(pathname: str, configuration: str, products: List[str] = None):
    """
    Create the smallest project file as possible on the disk.

    This method, which is for advanced usage, can be used to create initial projects
    for running scripts. Typically, you should use one of the SCADE Project Wizards
    to create projects for a given context, such as SCADE Suite or SCADE Test.

    Parameters
    ----------
    pathname : str
        Path of the project.
    configuration : str
        Name of the configuration. A project must have at least one configuration.
    products: List[str], default: None
        List of tags identifying the nature of the project.
        For example, ``SC`` indicates a SCADE Suite project.
    """
    if products is None:
        products = []
    f = open(pathname, 'w')

    f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    f.write('<Project id="1" oid_count="4" defaultConfiguration="2">\n')
    if len(products) > 0:
        f.write('\t<props>\n')
        f.write('\t\t<Prop id="3" name="@STUDIO:PRODUCT">\n')
        for product in products:
            f.write('\t\t\t<value>{0}</value>\n'.format(product))
        f.write('\t\t</Prop>\n')
        f.write('\t\t<Prop id="4" name="@SCADE:SAVEVERSION">\n')
        f.write('\t\t\t<value>SCADE65</value>\n')
        f.write('\t\t</Prop>\n')
        f.write('\t</props>\n')
    f.write('\t<configurations>\n')
    f.write('\t\t<Configuration id="2" name="{0}"/>\n'.format(configuration))
    f.write('\t</configurations>\n')
    f.write('</Project>\n')


# -----------------------------------------------------------------------------
# Helpers

# def ApplyPropSet(object, context, dict):
#     for att, value in dict.items():
#         try:
#             _scade_api.set(object, att, value)
#         except:
#             # can't raise an error since calling function
#             # is successful and shall return a value
#             print('warning: ' + context + ': Illegal attribute ' + att + '\n')

# class NoneParamError(Exception):
#     def __init__(self, context, name):
#         self.context = context
#         self.name = name
#     def __str__(self):
#         return self.context + ': ' + self.name + ': Illegal empty parameter'

# class NotObjectError(Exception):
#     def __init__(self, context, name, object):
#         self.context = context
#         self.name = name
#         self.object = object
#     def __str__(self):
#         return self.context + ': ' + self.name + ': Illegal parameter ' + str(self.object)

# class BadClassError(Exception):
#     def __init__(self, context, name, cls):
#         self.context = context
#         self.name = name
#         self.cls = cls
#     def __str__(self):
#         return self.context + ': ' + self.name + ': Illegal class ' + str(self.cls)


def _check_object(object_, context: str, name: str, classes: Tuple[Any, ...]):
    """Check the type of a parameter and raise a ``TypeError`` if it is not correct."""
    if not isinstance(object_, classes):
        cls = type(object_).__name__
        # classes is either a type or a tuple of types
        if isinstance(classes, tuple):
            types = ' or '.join([_.__name__ for _ in classes])
        else:
            types = classes.__name__
        message = '%s: %s: Illegal type %s, must be %s' % (context, name, cls, types)
        raise TypeError(message)


# to be moved to query.project
def _find_file_ref(project: std.Project, pathname: str) -> std.FileRef:
    """
    Search a project for a file with the provided path.

    Parameters
    ----------
    project : std.Project
        Input project.
    pathname : str
        Path of the project to search.

    Returns
    -------
    std.FileRef
    """
    path = abspath(pathname)
    for file_ref in project.file_refs:
        if path == abspath(file_ref.pathname):
            return file_ref
    return None


def _get_elements(parent: Union[std.Project, std.Folder]) -> List[std.Element]:
    """
    Get the contained elements of a project or a folder.

    Parameters
    ----------
    parent: Union[std.Project, std.Folder]
        Input project or folder.

    Returns
    -------
    List[std.Element]
    """
    return parent.roots if isinstance(parent, std.Project) else parent.elements


# def FindObject(context: object, name: str, role: str, att):
#     for item in _scade_api.get(context, role):
#         if _scade_api.get(item, att) == name:
#             return item
#     return None
